import pandas as pd
import numpy as np
from scipy.optimize import curve_fit

def fit_parameter(df: pd.DataFrame, parameter):
    """Fit the S curve of concentration-response
    
    Args:
        df: Dataframe from the dataset
        ax: Axes object from the matplotlib

    Returns:
        popt: parameters of the S curve
        perr: RMSE of the fitted curve
    
    Raise:
        RuntimeError: When the curve cannot be fitted
    """
    df = df.copy()
    freqs = df[parameter]
    concentrations = df.concentration
    hillcurve = HillCurve(concentrations, freqs)
    popt = hillcurve.popt
    perr = hillcurve.calc_perr()
    return popt, perr, hillcurve

def fsigmoid(x, a, x0, k, b):
    return k / (1.0 + np.exp(-a*(x-x0))) + b

class HillCurve:
    # Learn from https://github.com/jbloomlab/neutcurve/blob/master/neutcurve/hillcurve.py
    def __init__(self,
                 concentrations: np.ndarray,
                 responses: np.ndarray,
                 concentration_unit=-6,
                 boundary='auto',
                 logc=True):
        if logc:
            concentrations = concentrations.apply(np.log10) + concentration_unit
        if boundary == 'auto':
            c_min = concentrations.min()
            c_max = concentrations.max()
        else:
            c_max, c_min = boundary
        p_min = responses.min()
        p = responses - p_min
        p_diff = p.max()
        p = p / p_diff
        popt, pcov = curve_fit(fsigmoid, concentrations, p,
            method='dogbox', bounds=([1., c_min, -1, 0], [10, c_max, 1, 1]))
        self.popt = popt
        self.pcov = pcov
        self.p_min = p_min
        self.p_diff = p_diff
        self.c_max = c_max
        self.c_min = c_min
        self.concentrations = concentrations
        self.responses = responses
    
    @property
    def EC50(self):
        return self.popt[1]

    @property
    def curve_diff(self):
        return np.abs(self.popt[2]) * self.p_diff

    @property
    def hill(self):
        return self.popt[0] * np.sign(self.popt[3])

    def calc_perr(self):
        perr = np.sqrt(np.diag(self.pcov))
        return perr

    def predict(self, x):
        y = fsigmoid(x, *self.popt)
        y = y * self.p_diff + self.p_min
        return y
        


import pandas as pd
from cdwave import fnc


class Dataset:
    _data = None

    def __init__(self):
        if Dataset._data is None:
            Dataset._data = self.load_data()
        self.data = Dataset._data

    def get_series(self, dataname):
        df = self.data[dataname]
        series = df.set_index('time').iloc[:, 0]
        return series

    def get_waveform(self, dataname):
        series = self.get_series(dataname)
        wave = fnc.Waveform(series)
        return wave

    @staticmethod
    def load_data() -> dict:
        times = []
        dataname = ''
        waveforms = {}
        with open('tests/test_data.csv') as fp:
            for i, line in enumerate(fp):
                items = line.strip().split(',')
                if i % 2 == 0:
                    times = items[1:]
                    dataname = items[0]
                if i % 2 == 1:
                    waveforms[dataname] = pd.DataFrame({
                        'time': times,
                        items[0]: items[1:]
                    }, dtype=float)
        return waveforms

    def __getitem__(self, dataname):
        return self.data[dataname]



import os
import pickle
import gzip
import logging
import warnings
import pandas as pd
from cdwave import fnc

logger = logging.getLogger(__name__)
crt_path = os.path.abspath(os.path.dirname(__file__))

class WaveformFull:
    """WaveformFull contains all the information such as compound name,
    concentration or vendor, and all calculated parameters of a waveform.

    Attributes:
        profile: A dictionary of The profile of the waveform, including:
            plate: The plate name form which the waveform is generated.
            compound: The compound name of the waveform.
            concentration: The concentration of the compound of the waveform.
            well: The well of the waveform.
            cpid: Compound id.
        signal (dict): A dictionary with with keys, x and y, which are are time
            and calcium transient of the waveform.
        state: The state of the waveform, such as before treatment and after
            treatment.
        parameters (dict): A dictionary containing all the parameters of the
            waveform.

    Args:
        item (dict): A dictionary contains all the information of the waveform
            required keys: signals, plate, well, concentration. item['singal']
            is a dictionary with times (x) and signals (y)
            `{'x': [0.1, 0.2], 'y':[100, 101]}`

    Example:
        >>> item = {'signal': {'x': [0.1, 0.2], 'y': [100, 101]},
        ...         'plate': 'P1',
        ...         'compound': 'cmp1',
        ...         'concentration': 0.01,
        ...         'well': 'A1'}
        >>> wave = WaveformFull(item)
    """
    parameter_df = pd.read_csv(os.path.join(crt_path, 'parameters.csv'), index_col=0)
    parameter_names = parameter_df.index.tolist()
    parameter_annotations = parameter_df['Annotation'].to_dict()

    def __init__(self, item):
        if 'compound' not in item:
            item['compound'] = 'undefined'
        if 'concentration' not in item:
            item['concentration'] = 0
        if 'cpid' not in item:
            item['cpid'] = item['compound']
        self.signal = self.standardise_signal(item['signal'])
        self.parameters = {x: 0 for x in self.parameter_names}
        self.profile = {
            'plate': str(item['plate']),
            'compound': item['compound'],
            'concentration': item['concentration'],
            'well': item['well'],
            'vendor': item.get('vendor', 'None'),
            'cpid': str(item['cpid']),
            'state': item.get('state', 'treat')
        }

    def get_dict(self):
        d = {'signal': self.signal}
        d.update(self.profile)
        d['waveform'] = self
        return d

    @staticmethod
    def standardise_signal(signal: dict):
        signals = signal['y']
        if len(signals) < 5:
            warnings.warn(
                "Number of signal is less than 5 and cannot be smoothed")
            return signals
        signals = fnc.signal_filter(signals)
        minimum = signals.min()
        signal['y'] = signals - minimum
        return signal

    def get_signal_series(self):
        se = pd.Series(self.signal['y'], index=self.signal['x'])
        se.name = self.profile['well']
        se.index.name = 'time'
        return se

    def get_parameters(self, fillna=0):
        """Return a dictionary with all parameters
        Args:
            fillna: If fill na is 'raise', then an exception will be raised when
            a parameter has not been calculated. For other values, it will be used
            to fill the empty parameter.

        Returns:
            dict: A dictionary with all parameters listed in `parameter_names`
        """
        if fillna == 'raise':
            return {x: self.parameters[x] for x in self.parameter_names}
        return {x: self.parameters[x] if x in self.parameters else fillna for x in self.parameter_names}

    def __str__(self):
        return "waveform: {}".format(self.profile)


class Dataset:
    """A dataset contains a list of WaveformFull objects and a meta table
    with the information and parameters of the waveforms

    Attributes:
        waveforms: A list of WaveformFull objects.
        dataframe: The meta table of the waveforms.
        filtered_df: A view of the meta table.
        filterable_columns: columns that can be used to filter the waveforms
        size: Number of waveforms in the dataset

    Args:
        waveforms: A list of WaveformFull objects
    """

    def __init__(self, waveforms: list = None):
        self.waveforms = waveforms
        self._dataframe = None
        self.filtered_df = None
        self.filterable_columns = [
            'plate', 'compound', 'state', 'well', 'vendor', 'cpid', 'concentration']
        self.size = len(waveforms)

    def __len__(self):
        return len(self.waveforms)

    @staticmethod
    def loaddata(filepath) -> 'Dataset':
        file_items = filepath.split('.')
        if len(file_items) > 1 and file_items[-1] == 'gz' and file_items[-2] == 'pickle':
            fp = gzip.open(filepath, 'rb')
        elif file_items[-1] == 'pickle':
            fp = open(filepath, 'rb')
        else:
            return None
        dataset = pickle.load(fp)
        dataset.dataframe = dataset.get_df()
        dataset.filtered_df = dataset.dataframe
        fp.close()
        return dataset

    @staticmethod
    def concat(datasets: list):
        merged_dataset = datasets[0].copy()
        for dataset in datasets[1:]:
            merged_dataset.merge(dataset)
        return merged_dataset

    def merge(self, dataset):
        self.waveforms += dataset.waveforms

    def copy(self):
        return Dataset(self.waveforms)

    def get_df(self) -> pd.DataFrame:
        return pd.DataFrame([x.get_dict() for x in self.waveforms])

    def get_parameter_df(self) -> pd.DataFrame:
        """Return a dataframe with all papameters for each wave"""
        df = self.df[self.filterable_columns+['waveform']]
        for parameter in WaveformFull.parameter_names:
            df[parameter] = [x.parameters[parameter] for x in df.waveform]
        for p in ['noise', 'double_peak', 'uniform']:
            df[p] = df[p].astype(bool)
        return df

    def save(self, filename, compress=True):
        if compress:
            fp = gzip.open(filename + '.gz', 'wb')
        else:
            fp = open(filename, 'wb')
        pickle.dump(self, fp)
        fp.close()

    @property
    def df(self):
        if self._dataframe is None:
            self._dataframe = self.get_df()
            self.filtered_df = self._dataframe
        return self._dataframe

    @property
    def dtypes(self):
        return self.df.dtypes

    def filter_by_filters(self, filters, replace=True):
        """
        Args:
            filters: a dictionary of {column: value}
        return:
            filtered dataframe
        """
        df = self.df
        condition = [True] * len(df)
        for column in filters:
            value = filters[column]
            condition &= df[column] == value
        if replace:
            self.filtered_df = df[condition]
            return self.filtered_df
        return df[condition]


class DataLoader:

    def __init__(self, filepath: str = None, log=None):
        self.filepath = filepath
        self.dataframe = pd.DataFrame()
        self.signals = []
        self.waveforms = []
        self.pbar = None
        if filepath and not os.path.exists(filepath):
            raise Exception(
                "Cannot open the file {}, doesn't exist!".format(filepath))
        if log is None:
            self.log = logger.debug
        else:
            self.log = log

    def transfer(self) -> Dataset:
        """Parse the waveform from tables and generate a dataset

        Returns:
            Dataset: The dataset containing all the waveforms and parameters
        """
        self.log('Start parsing....')
        self.parse()
        self.log('Parsing complete, {} waveforms in total'.format(
            len(self.waveforms)))
        dataset = Dataset(self.waveforms)
        return dataset

    def parse(self):
        """This function should parse waveform data from files into self.waveforms"""
        raise NotImplementedError


class StandardCSVLoader(DataLoader):
    """A loader parsing "standard csv file". The format of the table file is like
    ```
    compound,concentration,well,plate,time,signal
    CP1,0.1,A1,P1,0,1000
    CP1,0.1,A1,P1,0.33,1001
    CP2,0.1,A2,P1,0,1000
    ```
    """

    def __init__(self, filepath: str = None, data: pd.DataFrame = None, log=None):
        if isinstance(filepath, pd.DataFrame):
            data = filepath
            filepath = None
        super().__init__(filepath=filepath, log=log)
        if filepath is None:
            if data is None:
                raise ValueError('File path and data cannot both be None')
            self.data = data

    def parse(self):
        if self.filepath is None:
            df = self.data
        else:
            df = pd.read_csv(self.filepath)
        required_columns = ['plate', 'compound',
                            'concentration', 'well', 'time', 'signal']
        attribute_columns = ['plate', 'compound',
                             'concentration', 'well', 'cpid', 'vendor']
        miss_column = set(required_columns) - set(df.columns)
        if miss_column:
            raise KeyError('Column {} is missing'.format(miss_column))
        if df.duplicated(['plate', 'well', 'time']).any():
            raise ValueError('Find duplicacy in set')
        for _, gdf in df.groupby(['plate', 'well']):
            item = gdf.iloc[0]
            input_item = {x: item[x] for x in attribute_columns if x in item}
            signal = {'x': gdf['time'].values, 'y': gdf['signal'].values}
            input_item['signal'] = signal
            waveform = WaveformFull(input_item)
            self.waveforms.append(waveform)


if __name__ == "__main__":
    print('Number of parameters: {}'.format(len(WaveformFull.parameter_names)))



from itertools import product
import pandas as pd
from cdwave import param

model_parameters = ['up_length', 'down_length', 'full_down', 'rd_ratio', 'freq',
                    'avg_amplitude', 'avg_lambda', 'tail_proportion', 'avg_tail', 'std_tail',
                    'avg_shoulder', 'avg_shoulder_tail', 'std_shoulder',
                    'std_shoulder_tail', 'std_amplitude', 'std_lambda', 'PW10_mean', 'PW10_std',
                    'PW25_mean', 'PW25_std', 'PW50_mean', 'PW50_std', 'PW80_mean',
                    'PW80_std', 'PW90_mean', 'PW90_std']


def four_point_parameter_generator(parameters, suffixes):
    param_matrix = {'parameter': [], 'suffix': [], 'product': []}
    suffixes = ['_slope', '_median', '_max']
    for p, s in product(parameters, suffixes):
        param_matrix['parameter'].append(p)
        param_matrix['suffix'].append(s)
        param_matrix['product'].append(p + s)
    param_factor_df = pd.DataFrame(param_matrix).set_index('product')
    return param_factor_df


def prepare_four_point_model(agg_df: pd.DataFrame, endpoint: pd.Series, parameters, suffixes):
    res = {}
    cmps = agg_df['cmp'].unique()
    for p in parameters:
        for suffix in ['_min', '_median', '_max']:
            res[p+suffix] = []
        res[p+'_slope'] = []
        for cmp in cmps:
            df = agg_df[agg_df['compound'] == cmp]
            k = param.npoint_descriptor(df, p, 3)
            for i, suffix in enumerate(suffixes):
                res[p+suffix].append(k[i])
            k, _ = param.linear_regression_with_logc(df, p, error=0)
            res[p+'_slope'].append(k)
    kdf = pd.DataFrame(res)
    kdf['endpoint'] = kdf['compound'].map(endpoint)
    features = four_point_parameter_generator(
        parameters, suffixes).index.values
    X = kdf[features]
    y = kdf['ednpoint']
    return X, y, kdf



import logging
import pandas as pd
import numpy as np
from scipy.stats import kstest
from scipy.signal import find_peaks
from scipy.signal import peak_widths
from scipy.interpolate import UnivariateSpline
from scipy.signal import butter
from scipy.signal import filtfilt
from scipy.signal import hilbert
from scipy.signal import savgol_filter
from scipy.signal import welch
import statsmodels.nonparametric.api as smnp
logger = logging.getLogger(__name__)


class Waveform:
    """Class for waveform analysis

    Args:
        series: A series of which the name is the well name, values are amplitudes

    Attributes:
        df (pd.DataFrame): The main dataframe of the waveform, including several important
            columns peak: 1: main peak, 2-x: double peaks; status: 0: normal,
            1: raising point, 2: peak, 3: down starting;
        num_peak (int): Number of peaks
        n (int): Number of points
        maximum (int): Maximum of amplitudes
        minimum (int): Minimum of amplitudes
    """
    max_shoulder_tail_ratio = 2.5

    def __init__(self, series):
        self.series = series
        self.name = series.name
        self.num_peak = 0
        self.df = pd.DataFrame(series).reset_index()
        self.df['status'] = 0
        self.n = len(self.df)
        self.maximum = self.series.max()
        self.minimum = self.series.min()
        self.variance = (self.maximum - self.minimum) / 10
        self.group_valley = {}
        self.fail_analysis = False
        self._group = None

    @property
    def group(self):
        """Return a iterator of group (i, gdf),excluding the first and the last"""
        def iterator():
            for i, gdf in self._group:
                if i in (self.num_peak, 0):
                    continue
                yield i, gdf
        if self._group is None:
            if self.num_peak == 0:
                raise ValueError("Cannot get group before getting peak")
            self._group = self.df.groupby('group')
        return iterator()

    def regroup(self):
        """After the peaks are changed, the groups need to be recalculated
        This function calculate the number of peaks and re-define the group
        number of each point
        """
        df = self.df
        last_point = 0
        peaks = df[df['peak'] == 1]['peak']
        self.num_peak = len(peaks)
        for i, point in enumerate(peaks.index):
            self.df.loc[last_point:point, 'group'] = i
            last_point = point
        self.df.loc[last_point:, 'group'] = self.num_peak
        group = df.groupby('group')
        self._group = group
        for i, gdf in group:
            cond = gdf['peak'].notna() & (gdf['peak'] > 1)
            n_subpeak = cond.sum()
            index = gdf.loc[cond].index
            self.df.loc[index, 'peak'] = list(range(2, n_subpeak+2))

    def analyse_normal_waves(self, diff_n=1, l=3):
        """Get status of points for a normal wave

        Args:
            series: A series of which the name is the well name, values are amplitudes
            diff_n: n times of continuous points higher than this point to be
                regarded as a starting point
            l: length of exploration to find a point after higher than
                upper_half, indicating it can be a starting point
        """
        series = self.series
        # 1 Starting point of raising, 2: Peak point, 0: Normal (down)
        last_status = 0
        status = []
        diff = series.diff(diff_n)
        diff.iloc[0] = 0
        diff.iloc[1] = series.iloc[1] - series.iloc[0]
        upper_half = series.max() / 2
        for i, row in enumerate(
                self.df[['category', 'peak', 'group', self.name]].itertuples(index=False)):
            if i >= self.n - l:  # For the last group, don't calculte status
                status.append(0)
                continue
            if last_status == 0:  # Checking if it starts to raise
                for j in range(max(1, diff_n - i), l+1):
                    # One point between i and i+diff_n is smaller than i, reject the starting point
                    if diff.iloc[i+j] < 0:
                        status.append(0)
                        break
                    # Once a point between i and i+l is higher than upper_half,
                    # accept the starting point.
                    if series.iloc[i+j] > upper_half:
                        status.append(1)
                        last_status = 1
                        break
                else:
                    status.append(0)
            elif last_status < 2 and row[1] == 1:
                last_status = 2
                status.append(2)
            elif last_status == 2 and row[2] in self.group_valley and \
                    row[3] - self.group_valley[row[2]] < self.variance:
                status.append(3)
                last_status = 0
            else:
                status.append(0)
        self.df['status'] = status

    def analyse(self):
        """Analyse the status of each point
        The status indicates the thrend of the point, such as rasing and
        declining. For waveforms of which frequency is high than 10, it will
        identify double peaks via their prominences. This function also
        calculates the valley possitions.
        """
        self.get_valleys()
        # Check if there is less then 10 points between two peaks
        df = self.df.reset_index()
        diff = df[df['prominence'] != 0]['index'].diff().dropna()
        high_freq = diff.max() < 10
        if high_freq:
            peaks = self.df[self.df['prominence'] != 0].index
            valley_pos = self.df[self.df['valley'] == 1].index
            self.df.loc[valley_pos, 'status'] = 1
            self.df.loc[peaks, 'status'] = 2
        else:
            self.fix_double_peak_by_prominence()
            self.analyse_normal_waves()
            self.fix_double_peak_by_tail()
            self.analyse_normal_waves()

    def check_status_points(self):
        """A debugging function checking whether any period has more than 1 or
        has not starting point or downing point.
        """
        Success = True
        for i, gdf in self.df.groupby('group'):
            if i in (self.num_peak, 0):
                continue
            starting_point = gdf[gdf['status'] == 1]
            downing_point = gdf[gdf['status'] == 3]
            if len(starting_point) > 1:
                logger.debug("More than one starting point in group %d", i)
                Success = False
            elif len(starting_point) == 0:
                logger.debug("No starting point in group %d", i)
                Success = False
            if len(downing_point) > 1:
                logger.debug("More than one downing_point in group %d", i)
                Success = False
            elif len(downing_point) == 0:
                logger.debug("No downing point in group %d", i)
                Success = False
        return Success

    def get_peaks(self, height=None, prominence=None):
        """Identify the peaks and group of the whole waveform

        Args:
            height: number or ndarray or sequence, optional Required height of peaks.
                Either a number, None, an array matching x or a 2-element sequence of the former.
                The first element is always interpreted as the minimal and the second,
                if supplied, as the maximal required height.
            prominence: number or ndarray or sequence, optional Required prominence of peaks.
                Either a number, None, an array matching x or a 2-element sequence of the former.
                The first element is always interpreted as the minimal and the second, if supplied,
                as the maximal required prominence. The default value is the maximum of 20 and 10%
                of maximum amplitude.
        """
        series = self.df[self.name]
        self.df['category'] = (series - self.minimum) // self.variance
        # opt_series is to prioritise early points
        opt_series = series - series.index * 0.0001
        if prominence is None:
            prominence = max(20, 0.1*self.maximum)
        peaks, properties = find_peaks(
            opt_series, height=height, prominence=prominence)
        prominences = properties['prominences']
        self.df['peak'] = 0
        self.df['prominence'] = 0
        for i in range(len(peaks)):
            peak = peaks[i]
            p = prominences[i]
            if i == 0:
                p = series[peak] - series[properties['right_bases'][i]]
            elif i == len(peaks)-1:
                p = series[peak] - series[properties['left_bases'][i]]
            self.df.loc[peak, 'peak'] = 1
            self.df.loc[peak, 'prominence'] = p
        self.num_peak = len(self.df[self.df['peak'] == 1])
        if self.num_peak == 0:
            return False
        self.regroup()
        return True

    def get_valleys(self):
        """Add valley infomation into the main dataframe"""
        self.df['valley'] = 0
        self.group_valley = {}
        for group_id, gdf in self.group:
            s = gdf[self.name]
            pos = s.idxmin()
            self.group_valley[group_id] = s[pos]
            self.df.loc[pos, 'valley'] = 1

    def fix_double_peak_by_prominence(self):
        """Identify double peaks by their prominence

        Definition of a double peak: A peak of which the prominence is below the
        threshold (see below about the threshold) and the signal value is close
        to the last real peak. Close means the difference between the signal
        value is smaller than variance (10% of maximum value).
        For waves of which the maximum is higher than 250, the threshold of the
        prominence is 0.7 * maixmum of all prominence. For others less than 250,
        the threshold is 0.5 * maximum of all prominence.
        Returns:
            Always True
        """
        if self.maximum < 250:
            prominences_t = 0.5
        else:
            prominences_t = 0.7
        # The order of current peak (1 is main peak, 2 is the first double peak)
        series = self.df[self.name]
        real_peaks = []
        peaks = self.df[self.df['peak'] == 1].index
        last_peak = None
        last_double_peak = None
        t = prominences_t * self.df.prominence.max()
        for peak in peaks:
            p = self.df.loc[peak, 'prominence']
            if p >= t:
                real_peaks.append(peak)
                self.df.loc[peak, 'peak'] = 1
                if last_double_peak and series[peak] - series[last_double_peak] < self.variance:
                    self.df.loc[last_double_peak, 'peak'] = 2
                last_peak = peak
            elif p < 0.3 * self.df.prominence.max():
                self.df.loc[peak, 'peak'] = 0  # Remove the peak
                continue
            elif last_peak and series[last_peak] - series[peak] < self.variance:
                self.df.loc[peak, 'peak'] = 2
                last_double_peak = None
            else:
                last_double_peak = peak
                self.df.loc[peak, 'peak'] = 0
        self.regroup()
        self.get_valleys()
        return True

    def calc_frequency_parameters(self):
        df = self.df
        lambdas = df[df['peak'] == 1].diff().iloc[1:]['time']
        avg_lambda = lambdas.mean() if len(lambdas) > 0 else 0
        std_lambda = lambdas.std() if len(lambdas) > 1 else 0
        if len(lambdas) > 5:
            quartiles = np.percentile(lambdas, [25, 50, 75])
            iqr = quartiles[-1] - quartiles[0]
            top = 1.5 * iqr + quartiles[-1]
            bottom = quartiles[0] - 1.5 * iqr
            inner_lambdas = lambdas[(lambdas <= top) & (lambdas >= bottom)]
            avg_inner_lambda = inner_lambdas.mean()
            std_inner_lambda = inner_lambdas.std()
        else:
            avg_inner_lambda = avg_lambda
            std_inner_lambda = std_lambda
        freq = (1 / avg_lambda * 100) if avg_lambda != 0 else 0
        return {
            'avg_lambda': avg_lambda,
            'std_lambda': std_lambda,
            'avg_inner_lambda': avg_inner_lambda,
            'std_inner_lambda': std_inner_lambda,
            'freq': freq
        }

    def calc_status_parameter(self):
        ups = []
        downs = []
        full_downs = []
        rd_ratios = []  # Raise decay ratio
        tails = []
        t_proportion = []  # Tail proportion
        for _, gdf in self.group:
            peak_time = gdf.iloc[0]['time']
            try:
                starting_point = gdf[gdf['status'] == 1].iloc[0]['time']
                downing_point = gdf[gdf['status'] == 3].iloc[0]['time']
            except IndexError:
                self.fail_analysis = True
                continue
            # gdf.index[-1] + 1 is next peak index
            next_peak_time = self.df.iloc[gdf.index[-1] + 1]['time']
            up = next_peak_time - starting_point
            down = downing_point - peak_time
            full_down = starting_point - peak_time
            tail = starting_point - downing_point
            ups.append(up)
            downs.append(down)
            full_downs.append(full_down)
            tails.append(tail)
            rd_ratios.append(up / full_down)
            t_proportion.append(tail / (up + full_down))
        if tails:
            avg_tail = np.mean(tails)
            std_tail = np.std(tails)
            t_proportion = np.mean(t_proportion)
        else:
            avg_tail = std_tail = t_proportion = 0
        return {
            'up_length': np.mean(ups) if ups else 0,
            'down_length': np.mean(downs) if downs else 0,
            'full_down': np.mean(full_downs) if full_downs else 0,
            'rd_ratio': np.mean(rd_ratios) if rd_ratios else 0,
            'avg_tail': avg_tail, 'std_tail': std_tail, 'tail_proportion': t_proportion
        }

    def calc_shoulder_parameters(self):
        """Calculate shoulder related parameters

        Shoulder parameters include mean and standard deviation of
        shoulder position (in amplitude), and median and standard deviation of
        shoulder_tail ratio. Median is used because sometime the ratio is
        unstable. For example, when tail is missed in one period, the ratio
        will become 100, which is cover other normal values.
        """
        shoulder_amplitudes = []
        shoulder_tail_ratio = []
        shoulder_position = []
        for _, gdf in self.group:
            amplitude = gdf.iloc[0][self.name]
            wdist = self.get_width_distribution(gdf)
            if not wdist:
                continue
            grid, y = wdist
            wd_peaks, wd_properties = find_peaks(
                y, prominence=0)  # Width Distribution
            peak_amplitudes = grid[wd_peaks]
            prominences = wd_properties['prominences']
            wd_tail = 0
            wd_shoulder_prominence = 0
            wd_shoulder = 0
            for wd_peak, prominence in zip(peak_amplitudes, prominences):
                if wd_peak > 0.80 * amplitude:
                    wd_tail = max(wd_tail, prominence)
                elif wd_peak > 0.15 * amplitude:
                    if prominence > wd_shoulder_prominence:
                        wd_shoulder_prominence = prominence
                        wd_shoulder = wd_peak
            shoulder_amplitudes.append(wd_shoulder)
            shoulder_position.append(wd_shoulder/amplitude)
            if wd_tail == 0 and wd_shoulder != 0:
                shoulder_tail_ratio.append(self.max_shoulder_tail_ratio)
            elif wd_tail != 0:
                shoulder_tail_ratio.append(
                    min(self.max_shoulder_tail_ratio, wd_shoulder_prominence/wd_tail))
            else:
                shoulder_tail_ratio.append(0)
        return {
            'avg_shoulder': np.mean(shoulder_position) if shoulder_position else 0,
            'avg_shoulder_tail': np.median(shoulder_tail_ratio) if shoulder_tail_ratio else 0,
            'std_shoulder': np.std(shoulder_position) if shoulder_amplitudes else 0,
            'std_shoulder_tail': np.std(shoulder_tail_ratio) if shoulder_tail_ratio else 0,
            'avg_shoulder_amp': np.mean(shoulder_amplitudes) if shoulder_amplitudes else 0,
            'std_shoulder_amp': np.std(shoulder_amplitudes) if shoulder_amplitudes else 0,
        }

    def calc_peak_width(self):
        df = self.df
        peaks = df[df['peak'] == 1].index
        rel_height_dict = {
            'PW10': 0.1,
            'PW25': 0.25,
            'PW50': 0.5,
            'PW80': 0.8,
            'PW90': 0.9
        }
        parameters = {name + '_mean': 0 for name in rel_height_dict}
        parameters.update({name + '_std': 0 for name in rel_height_dict})
        for name, rel_height in rel_height_dict.items():
            widths = peak_widths(df[self.name], peaks, rel_height)[0]
            parameters[name+'_mean'] = widths.mean()
            parameters[name+'_std'] = widths.std()
        return parameters

    def calc_amplitudes(self):
        df = self.df
        intensities = df.loc[df['peak'] == 1, self.name].values
        valleys = list(self.group_valley.values())
        amplitudes = []
        rms = np.sqrt(np.mean(np.square(df[self.name].values)))
        for i, v in enumerate(valleys):
            amplitudes.append(intensities[i] - v)
            amplitudes.append(intensities[i+1] - v)
        amplitudes = np.array(amplitudes)
        parameters = {
            'avg_amplitude': amplitudes.mean() if valleys else 0,
            'std_amplitude': amplitudes.std() if valleys else 0,
            'avg_intensity': intensities.mean(),
            'std_intensity': intensities.std(),
            'avg_valley': np.mean(valleys) if valleys else 0,
            'rms': rms
        }
        return parameters

    def calc_fft_freq_ratio(self):
        r"""Calculate the energy ratio between double frequency (minor) and major

        .. math::

            FFT Ratio = \frac{\sum_{x=f_{mi}-0.05}^{f_{mi}+0.05}p(x)}
                        {\sum_{x=f_{ma}-0.05}^{f_{ma}+0.05}p(x)}
        """
        signals = self.resample(100)
        frq, psd = wave_transform(signals)
        # Remove peaks of which frq < 0.1 Sometimes butter doesn't work well
        sorted_frq_psd = sorted(
            filter(lambda x: x[0] > 0.05, zip(frq, psd)),
            key=lambda x: x[1], reverse=True)
        # In some cases, double freq may have higher density than main freq
        # We should select the smaller one (ideally we should select the one
        # closer to real peak freqency)
        if sorted_frq_psd[0][1] / sorted_frq_psd[1][1] < 2 and \
                sorted_frq_psd[0][0] > sorted_frq_psd[1][0]:
            # Swap the first two
            sorted_frq_psd[0], sorted_frq_psd[1] = sorted_frq_psd[1], sorted_frq_psd[0]
        main_frq, main_psd = sorted_frq_psd[0]
        if main_frq == 0 or main_psd == 0:
            return 0
        for _frq, _psd in sorted_frq_psd[1:]:
            if 1.9 < _frq/main_frq < 2.1:
                minor_frq = _frq
                break
        else:
            return 0
        S_main_fft = psd[(main_frq-0.05 < frq) & (frq < main_frq+0.05)]
        P_main = np.sum(S_main_fft) * (frq[1] - frq[0])
        S_minor_fft = psd[(minor_frq-0.05 < frq) & (frq < minor_frq+0.05)]
        P_minor = np.sum(S_minor_fft) * (frq[1] - frq[0])
        if P_minor == 0:
            return 0
        return P_minor / P_main

    def get_parameters(self):
        """Get all parameters from the waveform"""
        df = self.df
        is_double_peak = (df['peak'] > 1).any()
        max_combo_peaks = df['peak'].max()
        if self.num_peak < 4:
            ks_p = 1
        elif self.num_peak < 6:
            ks_p = self.peak_uniform_test(interpolation=True)
        else:
            ks_p = self.peak_uniform_test()
        parameters = {'maximum': self.maximum, 'double_peak': is_double_peak,
                      'uniform': ks_p > 0.99, 'noise': self._noise_test(),
                      'n_peak': self.num_peak, 'max_combo_peaks': max_combo_peaks,
                      'fail_analysis': self.fail_analysis,
                      'fft_ratio': self.calc_fft_freq_ratio()
                      }
        parameters.update(self.calc_frequency_parameters())
        parameters.update(self.calc_status_parameter())
        parameters.update(self.calc_shoulder_parameters())
        parameters.update(self.calc_peak_width())
        parameters.update(self.calc_amplitudes())
        return parameters

    def draw_status(self, figure):
        """Plot the points into a figure colored by their status
        black indicates normal points; orange is starting point, the first point
        of raising wave; red is the peak and green is the first point of the
        tail.
        Args:
            figure: A Matplotlib figure object
        """
        figure.clear()
        colors = {0: 'black', 1: 'orange', 2: 'red', 3: 'green'}
        ax1 = figure.add_subplot(111)
        for color in colors:
            tdf = self.df[self.df['status'] == color]
            ax1.plot(tdf['time'], tdf[self.series.name],
                     'o', color=colors[color])
        tdf = self.df[(self.df['peak'] > 1)]
        ax1.plot(tdf['time'], tdf[self.series.name], 'mo')

    def draw_series(self, figure, series=None, style='-'):
        """Plot points of the waveform
        Args:
            figure: A Matplotlib figure object
            series: A pandas Series of the waveform
            style: A Maplotlib line style, `-` by default indicating a line plot
        """
        if series is None:
            series = self.series
        elif type(series) == str:
            series = self.df[series]
        figure.clear()
        ax1 = figure.add_subplot(111)
        ax1.plot(series.index, series, style)
        return ax1

    def export(self, filename):
        """Export the waveform to a csv file"""
        self.df.to_csv(filename)

    def peak_uniform_test(self, interpolation=False):
        """Test if the peak points are in uniform distribution

        Args:
            interpolation: True if interpolation is applied to the waveform.
                Mostly used when ther are only 3 peaks in the waveform so we
                need to interpolate the points between the three points.
                Otherwise KS test may underestimate the possibility

        Returns:
            float: The probability of of waveform being in uniform distribution
        """
        df = self.df
        x = df[df['peak'] == 1].index
        # First to check if the minimum and maximum is close to the boundary
        diff_max = np.diff(x).max()
        if x[0] > diff_max or x[-1] + diff_max < df.index.max():
            return 0
        x = x - min(x)
        x = x / max(x)
        if interpolation:
            new_x = [x[0]]
            for i in range(len(x)-1):
                new_x.append(x[i]+(x[i+1]-x[i])/2)
                new_x.append(x[i+1])
            x = new_x
        _, p = kstest(x, 'uniform')
        return p

    def _noise_test(self):
        '''(Deprecated) Test if this waveform is noise'''
        sum_error = 0
        for _, gdf in self.group:
            p1 = gdf.iloc[0][self.name]
            p2 = self.df.loc[gdf.index[-1] + 1, self.name]
            p_max = max(p1, p2)
            for x in gdf[self.name]:
                if x > p_max:
                    sum_error += 1
                if sum_error > 2:
                    return True
        return False

    def get_width_distribution(self, gdf: pd.DataFrame, normalise=True):
        """Calculate the distirbution of width between position of the peak point
        and the points.

        Args:
            gdf: subdataframe of a group

        Returns:
            tuple: A tuple of grid and y. `grid` is grid sampled in all widthes.
            `y` is the probability of the width
        """
        peak_amplitude = gdf.iloc[0][self.name]
        next_start = gdf[gdf['status'] == 1]
        if len(gdf) < 10 or len(next_start) == 0:
            return False
        next_start = next_start.index[0]
        prs = []
        for x in gdf.loc[:next_start, self.name]:
            pr = peak_amplitude - x
            prs.append(pr)
        kde = smnp.KDEUnivariate(prs)
        kde.fit('biw', bw=0.2*peak_amplitude, fft=False, gridsize=100,
                cut=3, clip=(-np.inf, np.inf))  # TODO hyper-parameter bw
        grid, y = kde.support, kde.density
        if normalise:
            y = y / y.max()  # Normalise the amplitude of density probability function
        return grid, y

    def fix_double_peak_by_tail(self, min_amplitude=100, std_threshold=0.5):
        """Identify double peaks by comparing the average signal values of tails

        In some situations, "real" peaks have a long tail but the subpeaks
        don't, so we can recognise the subpeaks by comparing their tail with
        maximum. But for some waveforms, the tails are too short to be used as a
        symbol to identify double peaks.

        Args:
            min_amplitude: Minimum of amplitude the wave should have as a
                prerequisite to find double peak.
            std_threshold: Number of times that the standard deviation of the
                tails is higher than mean of the tails which will indicate
                there is double peak.
        Returns:
            bool: If starting point or downing point were not found, return
                False and self.fail_analysis will be True. Otherwise always
                return True
        """
        if self.maximum <= min_amplitude:
            return False
        tails = []
        group = self.df.groupby('group')
        for i, gdf in group:
            if i in (0, self.num_peak):
                continue
            starting_point = gdf[gdf['status'] == 1]
            downing_point = gdf[gdf['status'] == 3]
            if len(downing_point) == 0:
                self.fail_analysis = True
                return False
            if len(starting_point) == 0:
                # This means starting_point is the same to downing point
                tails.append(0)
            else:
                tails.append(
                    starting_point.iloc[0]['time'] - downing_point.iloc[0]['time'])
        tails = np.array(tails)
        if len(tails) == 0:
            return False
        std_tail = tails.std()
        avg_tail = tails.mean()
        max_tail = tails.max()
        need_regroup = False
        if std_tail > std_threshold * avg_tail:
            for i, tail in enumerate(tails):
                if tail < max_tail - std_tail:
                    self.df.loc[group.get_group(i+1).index[0], 'peak'] = 2
                    need_regroup = True
        if need_regroup:
            self.regroup()
            self.get_valleys()  # TODO Maybe it can be merged to regroup
        return True

    def resample(self, sample_rate=100) -> np.ndarray:
        """Resample points in the waveform

        Args:
            sample_rate: How many points per second, default 100

        Returns:
            np.ndarray: An numpy array with `sample_rate` points per second
        """
        x = self.df['time']
        y = self.df[self.name]
        # new_y = savgol_filter(y, 5, 2)
        x_new = np.linspace(x.min(), x.max(), int(x.max()*sample_rate+1))
        interp = UnivariateSpline(x, y, k=3)
        return interp(x_new)

    def standardise_by_filter(self):
        signals = self.series.values
        new_signals = signal_filter(signals)
        series = pd.Series(
            new_signals, index=self.series.index, name=self.series.name)
        self.__init__(series)


def draw_multiple(figure, df, span=0):
    def make_slice(array):
        if span > 0:
            return array[:span]
        else:
            return array
    n = len(df)  # Only support 8 now
    if n > 8:
        logger.warning("Only the first 8 waveforms are shown")
        df = df.iloc[:8]
    figure.clear()
    axes = figure.subplots(2, 4, sharex='all', sharey='all')
    for i, row in df.reset_index().iterrows():
        ax = axes[i//4][i % 4]
        ax.plot(make_slice(row.signal['x']), make_slice(row.signal['y']), '-')
        ax.set_title('{:.3f}'.format(row['concentration']))


def wave_transform(signals: np.ndarray, sample_rate=100, method='fft'):
    """Transform the wave using methods such as Fourier transformation

    Args:
        signals (np.ndarray): resampled signals from waveform
        sample_rate (int): How many points per second, default 100
        method (str): Transformation method, default `fft`, fast fourier transform

    Returns:
        tuple: A tuple containing:
            - **frq** (*np.ndarray*): Frequency points (Hz)
            - **psd** (*np.ndarray*): Power Spectral Density from FFT
    """
    freq_range = np.array([0.01, 1])  # frequency from 0.03 to 1 Hz
    # Smooth by
    b, a = butter(2, freq_range/0.5/sample_rate, 'bandpass')
    signals = filtfilt(b, a, signals)
    if method == 'fft':
        datalen = len(signals)
        frq = np.fft.fftfreq(datalen, d=((1/sample_rate)))
        frq = frq[range(int(datalen/2))]
        Y = np.fft.fft(signals)/datalen
        Y = Y[range(int(datalen/2))]
        psd = np.power(np.abs(Y), 2)
    elif method == 'hilbert':
        frq = (np.arange(len(signals)) / sample_rate)[1:]
        analytic_signal = hilbert(signals)
        # amplitude_envelope = np.abs(analytic_signal)
        instantaneous_phase = np.unwrap(np.angle(analytic_signal))
        instantaneous_frequency = (np.diff(instantaneous_phase) /
                                   (2.0*np.pi) * sample_rate)
        psd = instantaneous_frequency
    elif method == 'welch':
        frq, psd = welch(signals, fs=100, nperseg=len(signals)-1)
    else:
        raise ValueError('The method is not Support')
    return frq, psd


def signal_filter(signals: np.ndarray, window_length=5) -> np.ndarray:
    """Apply a Savitzky-Golay filter to an array."""
    return savgol_filter(signals, window_length, 2)



import logging
import os
from multiprocessing import Pool
from collections.abc import Callable
from tqdm import tqdm
from joblib import Parallel, delayed
from cdwave.data import WaveformFull, Dataset
from cdwave.fnc import Waveform

logger = logging.getLogger(__name__)


def calc_parameter(waveform: WaveformFull) -> dict:
    """Calculate parameters of a waveform"""
    series = waveform.get_signal_series()
    wave = Waveform(series)
    if not wave.get_peaks():
        return None
    wave.analyse()
    try:
        r = wave.get_parameters()
    except Exception as e:
        logger.error('Cannot get parameters of %s', waveform)
        raise e
    return r


pbar_wrapper = {'x': None}


def default_process_fnc(status=-1, total=0):
    pbar = pbar_wrapper['x']
    if status == 0 and total != 0:
        pbar_wrapper['x'] = tqdm(total=total)
    elif status == 1:
        pbar.update()
    elif status == 2:
        pbar.close()


def calc_parameters_for_waveforms(dataset: Dataset,
                                  process_fnc: Callable = None,
                                  batch: int = 200,
                                  processes: int = None,
                                  custom_calculator: Callable = None):
    """Calculate parameter for waveforms

    Args:
        dataset: The waveform dataset
        process_fnc: A processing function used to send out the progress,
            see `default_process_fnc`, which uses tqdm
        batch: Batch size for multi-processing
        processes: Number of processors
        custom_calculator: A custom calculator which can setup the custom thresholds.
            If `None`, `calc_parameter` will be used by default.
    """
    if processes is None:
        try:
            processes = int(os.environ['NUMEXPR_MAX_THREADS'])
        except KeyError:
            processes = 10
        except ValueError:
            processes = 10
    if custom_calculator is None:
        custom_calculator = calc_parameter
    n = dataset.size
    if process_fnc:
        process_fnc(status=0, total=n)
    reses = []
    batch_waveforms = []
    for i, waveform in enumerate(dataset.waveforms):
        batch_waveforms.append(waveform)
        if i % batch == batch - 1 or i == n - 1:
            with Pool(processes=processes) as pool:
                for _waveform in batch_waveforms:
                    reses.append(pool.apply_async(
                        custom_calculator, (_waveform,)))
                for j, res in enumerate(reses):
                    r = res.get()
                    if r is not None:
                        batch_waveforms[j].parameters = r
            reses = []
            batch_waveforms = []
        if process_fnc:
            process_fnc(status=1)
    if process_fnc:
        process_fnc(status=2)


def calc_parameters_with_threshold(dataset: Dataset,
                                   threshold: int = 100,
                                   processes: int = None,
                                   custom_calculator: Callable = None):
    """Calculate parameter for waveforms

    Args:
        dataset: The waveform dataset
        process_fnc: A processing function used to send out the progress,
            see `default_process_fnc`, which uses tqdm
        batch: Batch size for multi-processing
        processes: Number of processors
        custom_calculator: A custom calculator which can setup the custom thresholds.
            If `None`, `calc_parameter` will be used by default.
    """
    if processes is None:
        try:
            processes = int(os.environ['NUMEXPR_MAX_THREADS'])
        except (KeyError, ValueError):
            processes = 4
    if custom_calculator is None:
        custom_calculator = calc_parameter_with_threshold
    logger.debug('Calculating parameters for %d waveforms with %d preceossors', len(dataset), processes)
    parameters = Parallel(processes)(
        delayed(custom_calculator)(_waveform, threshold) for _waveform in tqdm(dataset.waveforms))
    for waveform, p in zip(dataset.waveforms, parameters):
        if p is not None:
            waveform.parameters = p


def calc_parameter_with_threshold(waveform: WaveformFull, threshold) -> dict:
    """Calculate parameters of a waveform"""
    series = waveform.get_signal_series()
    wave = Waveform(series)
    if not wave.get_peaks(height=threshold):
        return None
    wave.analyse()
    try:
        r = wave.get_parameters()
    except Exception as e:
        logging.error('Cannot get parameters of %s', waveform)
        raise e
    return r



import unittest
from io import StringIO
import pandas as pd
from cdwave import data, fnc
from tests import Dataset


class TestData(unittest.TestCase):
    _data = Dataset()

    def get_mock_waveform(self):
        df = self._data['D1']
        input_item = {
            'plate': 'plate1',
            'compound': 'compound1',
            'concentration': 100,
            'cpid': 'x',
            'vendor': 'AstraZeneca',
            'well': 'B14',
            'state': 'prior'
        }
        input_item['signal'] = {
            'x': df['time'].values,
            'y': df['B14'].values
        }
        waveform = data.WaveformFull(input_item)
        return waveform

    def test_parameter_consistency(self):
        waveform = self.get_mock_waveform()
        series = waveform.get_signal_series()
        wave = fnc.Waveform(series)
        wave.get_peaks()
        wave.analyse()
        parameters = wave.get_parameters()
        parameters_more = set(parameters.keys()) - \
            set(data.WaveformFull.parameter_names)
        self.assertFalse(
            parameters_more, "These parameters are not included in default parameter names")
        default_more = set(data.WaveformFull.parameter_names) - \
            set(parameters.keys())
        self.assertFalse(default_more, "These parameters are not calculated")

    def test_csvloader(self):
        csv_string = """compound,concentration,well,plate,time,signal
        CP1,0.1,A1,P1,0,1000
        CP1,0.1,A1,P1,0.33,1001
        CP2,0.1,A2,P1,0,1000""".replace(' ', '')
        obj = StringIO(csv_string)
        df = pd.read_csv(obj)
        dataset = data.StandardCSVLoader(df).transfer()
        self.assertEqual(len(dataset), 2)
        obj = StringIO(csv_string+'\nCP1,0.1,A1,P1,0,100')
        df = pd.read_csv(obj)
        with self.assertRaises(ValueError):
            dataset = data.StandardCSVLoader(df).transfer()



import unittest
import numpy as np
import pandas as pd
from cdwave import param

parameters = ['up_length', 'down_length',
              'full_down', 'rd_ratio', 'freq', 'maximum']


def generate_random_params():
    df = pd.DataFrame({
        'plate': ['A', 'A', 'A', 'A', 'B', 'B', 'B', 'B'],
        'compound': ['NegCtrl', 'NegCtrl', 'A', 'A', 'NegCtrl', 'NegCtrl', 'A', 'A'],
        'well': ['A1', 'A1', 'A2', 'A2', 'B1', 'B1', 'B2', 'B2'],
        'state': ['prior', 'treat', 'prior', 'treat', 'prior', 'treat', 'prior', 'treat'],
        'concentration': [0, 0, 10, 5, 0, 0, 10, 5],
        'index': [1, 2, 3, 4, 5, 6, 7, 8],
        'vendor': ['AstraZeneca'] * 8
    })
    for p in parameters:
        df[p] = np.random.random(8)
    return df


class TestParam(unittest.TestCase):
    _data = generate_random_params()

    def test_normalise_baseline(self):
        sub_params = ['rd_ratio']
        div_params = ['up_length', 'down_length',
                      'full_down', 'freq', 'maximum']
        df = param.normalise_by_baseline(self._data, sub_params, div_params)
        self.assertTrue(df[parameters].notna().all().all())

    def test_normalise_negctrl(self):
        sub_params = ['rd_ratio']
        div_params = ['up_length', 'down_length',
                      'full_down', 'freq', 'maximum']
        data = self._data.query('state=="treat"')
        df = param.normalise_by_negctrl(
            data, standardisers={'sm': sub_params, 'sdm': div_params}, control_compound='NegCtrl')
        self.assertTrue(df[parameters].notna().all().all())



import logging
from sklearn.manifold import TSNE
from sklearn.decomposition import PCA
from sklearn.decomposition import FactorAnalysis as FA
from sklearn import preprocessing
from sklearn.linear_model import LinearRegression
from scipy.stats import median_absolute_deviation
import pandas as pd
import numpy as np
from tqdm import tqdm
from cdwave import data
from cdwave import hillcurve
logger = logging.getLogger(__name__)


def parameter_projection(df: pd.DataFrame, parameters: list = None, method='tsne', n_components=2):
    if parameters is None:
        parameters = data.WaveformFull.parameter_names
    X = df[parameters].astype('float').values  # Transfer bool to float
    if method == 'pca':
        X = preprocessing.scale(X)
    models = {
        'tsne': TSNE,
        'pca': PCA,
        'fa': FA
    }
    model = models[method](n_components)
    newX = model.fit_transform(X)
    for i in range(n_components):
        df['X%d' % (i+1)] = newX[:, i]
    return df, model


def aggrate_parameters(df: pd.DataFrame,
                       parameters: list = None, method="median",
                       compound_column='uniname',
                       plates: dict = None):
    """Aggregate the parameters of the same compound under the same concentration
    by methods such as median or mean

    Args:
        df: Dataframe of the whole parameters
        parameters: The parameter list to process
        method: The method to aggrate the parameters
        plates: A dictionary with the key of compounds and values of list of
            plates to use.

    Returns:
        DataFrame: Dataframe with aggregated parameters.
    """
    res = []
    if parameters is None:
        parameters = data.WaveformFull.parameter_names
    for (concentration, compound), gdf in df.groupby(['concentration', compound_column]):
        if plates and compound in plates:
            gdf = gdf[gdf.plate.isin(plates[compound])]
        if len(gdf) == 0:
            continue
        s = gdf[parameters].agg(method)
        s['compound'] = compound
        s['concentration'] = concentration
        res.append(s)
    return pd.DataFrame(res)


def parameter_correlation(df: pd.DataFrame, parameters: list = None):
    """Calculate the correlation between the parameters"""
    if parameters is None:
        parameters = data.WaveformFull.parameter_names
    X = df[parameters].astype(float).values
    corr = np.corrcoef(X, rowvar=False)
    return corr


def remove_well_by_baseline(pdf: pd.DataFrame):
    """Remove wells by baseline (prior)

    Args:
        pdf: DataFrame of one plate

    Returns:
        list: A list of removing wells
    """
    df = pdf[pdf['state'] == 'prior']
    wells = df[(df['max_combo_peaks'] > 1) |
               (df['std_lambda'] > 1) |
               (df['maximum'] < 100) |
               (df['fail_analysis'] == 'FALSE')]['well'].tolist()
    return wells


def calc_rcv(x: np.ndarray):
    """Robust coefficient of variation

    .. math::

       MAD = med | x_i - m |

       RCV_M = 1.4826 * \\frac{MAD}{m}

    Args:
        x: A 1-d array of parameters

    Returns:
        float: robust coefficient of variation
    """
    m = x.median()
    xm = x - m
    mad = xm.median()
    return 1.4826 * mad / m


def remove_low_quality(df: pd.DataFrame):
    """Remove waveforms with low quality
    The low quality includes:

    1. Double peak in negative control
    2. Low quality in baseline

    Args:
        df: A dataframe of all the samples with parameters

    Returns:
        tuple: A tuple containing a filtered dataframe and a dictionary of
            removed wells
    """
    pdfs = []
    removed_wells = {}
    for plate, pdf in df.groupby('plate'):
        removed_wells[plate] = {}
        removing_wells = remove_well_by_baseline(pdf)
        removed_wells[plate].update(
            {x: 'low baseline' for x in removing_wells})
        pdf = pdf[~pdf['well'].isin(removing_wells)]
        removing_wells = pdf[pdf['compound'].isin(['NegCtrl', 'DMSO']) &
                             ((pdf['max_combo_peaks'] > 1) | (pdf['std_lambda'] > 1))]['well'].tolist()
        removed_wells[plate].update({x: 'double peak' for x in removing_wells})
        pdf = pdf[~pdf['well'].isin(removing_wells)]
        rcv = calc_rcv(pdf[pdf['compound'].isin(['NegCtrl', 'DMSO'])]['freq'])
        if rcv >= 0.2:
            logger.debug(
                "Plate %s is removed due to the rcv of %s", plate, rcv)
            continue
        pdfs.append(pdf)
    return pd.concat(pdfs), removed_wells


def normalise_by_negctrl(df: pd.DataFrame,
                         standardiser: str = 'sdm',
                         standardisers: dict = None,
                         parameters: list = None,
                         control_compound: str = 'DMSO'):
    """Normalise the parameters by negative control of the plate
    Due to the fact that there will be one negative control in a plate, we
    use mean to aggragate the parameters.

    Args:
        df: DataFrame of parameters got from CardioWave
        standariser: Method to standardise the datak, including  
            sdm: Subtract and divide by median of negative control  
            sm: Subtract by median of negative control  
            smdmad: Subtract median and divide by median absolute deviation
        standardisers: A dictionary of which the keys are standarise methods
            and the values are parameters implementing the standardisers.
        control_compound: The name of control samples in the compound column

    Return:
        DataFrame: Normalised parameters
    """
    if parameters is None:
        parameters = data.WaveformFull.parameter_names
    # TODO remove specific names such as DMSO, NegCtrl
    control_dict = {'GSK': 'DMSO', 'AstraZeneca': 'NegCtrl'}
    agg_df_list = []
    for (plate, vendor), pdf in df.groupby(['plate', 'vendor']):
        if control_compound is None:
            control_name = control_dict[vendor]
        else:
            control_name = control_compound
        if standardisers is None:
            standardisers = {standardiser: parameters}
        control_samples = pdf[pdf['compound'] == control_name]
        assert pdf.well.duplicated().sum() == 0
        df_list = []
        for method in standardisers:
            samples = []
            ps = standardisers[method]
            if method == 'smdmad':
                mad = median_absolute_deviation(control_samples[ps])
            median = control_samples[ps].agg('median')
            for idx, item in pdf.iterrows():
                if method == 'sm':
                    sample = item[ps] - median
                elif method == 'sdm':
                    sample = (item[ps] - median).divide(median)
                elif method == 'smdmad':
                    sample = (item[ps] - median).divide(mad)
                else:
                    raise ValueError(f'Method "{method}" is invalid')
                samples.append(sample)
            df_list.append(pd.DataFrame(samples, index=pdf.index))
        tdf = pd.concat(df_list, axis=1)
        for col in ['compound', 'concentration', 'well', 'vendor', 'plate']:
            tdf[col] = pdf[col]
        agg_df_list.append(tdf)
    agg_df = pd.concat(agg_df_list)
    return agg_df


def normalise_by_baseline(df: pd.DataFrame,
                          subtract_params: list,
                          divide_params: list,
                          divide_only_params: list = None):
    """Normalise the parameters by baseline of the well

    Args:
        subtract_params: Parameter list to be subtracted only
        divide_params: Parameter list to be subtracted and divided
        divide_only_params: Parameter list to be divided only
    Return:
        DataFrame: Normalised parameters
    """
    if divide_only_params is None:
        divide_only_params = []
    if len(set(df.state) & set(['prior', 'treat'])) != 2:
        raise ValueError("states of treat and prior are required")
    agg_df_list = []
    for (plate, well), cdf in df.groupby(['plate', 'well']):
        baseline = cdf[cdf['state'] == 'prior']
        treat = cdf[cdf['state'] == 'treat']
        if len(baseline) != 1 or len(treat) != 1:
            continue
        baseline = baseline.iloc[0]
        treat = treat.iloc[0]
        sub_sample = treat[subtract_params] - baseline[subtract_params]
        div_sample = (treat[divide_params] - baseline[divide_params]).divide(
            baseline[divide_params])
        div_only_sample = treat[divide_only_params].divide(baseline[divide_only_params])
        sample_dict = {'compound': baseline['compound'], 'plate': plate,
                       'concentration': baseline['concentration'],
                       'well': well,
                       'vendor': baseline['vendor'],
                       'index': treat['index']}
        sample_dict.update(sub_sample.to_dict())
        sample_dict.update(div_sample.to_dict())
        sample_dict.update(div_only_sample.to_dict())
        agg_df_list.append(sample_dict)
    return pd.DataFrame(agg_df_list)


def select_concentration(df: pd.DataFrame, c: float, lim=(0.01, 10)):
    ids = []
    for (plate, compound), pdf in df.groupby(['plate', 'cmp']):
        tmp = pdf[pdf['concentration'] >= c].sort_values(by='concentration')
        if len(tmp) == 0:
            tmp = pdf.sort_values(by='concentration', ascending=False)
        if len(tmp) == 0:
            logger.warning('No sample in %s - %s', plate, compound)
        concentration = tmp.iloc[0]['concentration']
        if concentration / c > lim[1] or concentration / c < lim[0]:
            continue
        ids.append(tmp.index[0])
    return df.loc[ids]


def select_concentration_by_log(df: pd.DataFrame, c: float):
    ids = []
    for _, pdf in df.groupby(['plate', 'cmp']):
        tmp = pdf[['concentration']].copy()
        tmp['logc'] = np.log10(tmp['concentration'])
        tmp['diff_logc'] = np.abs(tmp['logc'] - np.log10(c))
        ids.append(tmp.sort_values(by='diff_logc').index[0])
    return df.loc[ids]


def calc_inflection(df: pd.DataFrame, parameters=None, min_n=4):
    if parameters is None:
        parameters = ['freq', 'rms', 'PW10_mean']
    items = []
    for (plate, cmp), cdf in tqdm(df.groupby(['plate', 'cmp'])):
        item = {'plate': plate, 'cmp': cmp}
        if len(cdf) <= min_n:
            continue
        for parameter in parameters:
            cs = cdf['concentration']
            fs = cdf[parameter]
            try:
                curve = hillcurve.HillCurve(cs, fs)
            except ValueError:
                print(plate, cmp)
                continue
            perr = curve.calc_perr()
            if perr[2] > np.abs(curve.popt[2]):
                # A magic number, to be improved
                item[f'{parameter}_EC50'] = -4.3
            else:
                item[f'{parameter}_EC50'] = curve.EC50
            # item[f'{parameter}_EC50_std'] = min(perr[1], 10)
            item[f'{parameter}_DIS'] = curve.curve_diff
            # item[f'{parameter}_DIS_std'] = min(perr[2], 10)
        items.append(item)
    return pd.DataFrame(items)


def select_plates(df: pd.DataFrame, t=0.2):
    """If the amplitude and freq of the lowest concentration is out of +- 0.2
    then remove the plate

    Args:
        df: input dataframe of the parameters
        t: threshold of the good quality

    Returns:
        dict: A dictionary with key of compounds and value of a list of available
        plates
    """
    res = {}
    for (plate, mol), cdf in df.groupby(['plate', 'uniname']):
        c_min = cdf['concentration'].min()
        tmp = cdf[(cdf['concentration'] == c_min) &
                  ((np.abs(cdf['avg_amplitude']) >= t) |
                   (np.abs(cdf['freq']) >= t))]
        if len(tmp) > 0:
            continue
        res.setdefault(mol, []).append(plate)
    return res


def linear_regression_with_logc(df: pd.DataFrame,
                                parameter: str,
                                remove_beatstop=True,
                                error='raise'):
    """Use linear regression to derive slope and intercept for a parameter

    Args:
        df: A dataframe that contains the samples needed.
        parameter: The name of the parameter that are included in the dataframe
        remove_beatstop: Whether to remove the samples of which 'beat_stop' is True
        error: The return k and b if all the samples are beat_stop or there is no sample.
            if `raise` is used, it will raise a ValueError from LinearRegression

    Returns:
        Tuple: slope (k) and intercept (b)
    """
    if parameter not in df or 'logc' not in df:
        raise KeyError("Do not have the parameter or logc")
    if remove_beatstop and 'beat_stop' not in df:
        raise KeyError("Do not have beat_stop column")
    if remove_beatstop:
        df = df[~df['beat_stop']]
    if error != 'raise' and len(df) == 0:
        return error, error
    reg = LinearRegression().fit(
        df['logc'].values.reshape(-1, 1), df[parameter])
    k = reg.coef_[0]
    b = reg.intercept_
    return k, b


def npoint_descriptor(df: pd.DataFrame, parameter: str, n: int):
    "This function only works for 8 concentrations"
    df = df.sort_values(by='concentration')
    res = []
    if n == 3:
        res.append(df.iloc[0][parameter])
        # res.append(df[parameter].median())
        res.append((df.iloc[3][parameter]+df.iloc[4][parameter])/2)
        res.append(df.iloc[len(df)-1][parameter])
    elif n == 8:
        res = df[parameter].values
        if len(res) != 8:
            logger.warning(
                'This compound does not have 8 points - %d', len(res))
    else:
        raise ValueError("Only 3 or 8 points are supported.")
    return res



import unittest
import warnings
import numpy as np
from cdwave import fnc
from tests import Dataset


class TestFnc(unittest.TestCase):
    data = Dataset()

    def test_find_peak(self):
        df = self.data['D1']
        series = df.set_index('time')['B14']
        wave = fnc.Waveform(series)
        self.assertTrue(wave.get_peaks(), "Cannot get peaks")
        wave.analyse()
        self.assertEqual(wave.num_peak, 17)

    def test_double_peak(self):
        df = self.data['double_peak']
        series = df.set_index('time')['F6']
        wave = fnc.Waveform(series)
        self.assertTrue(wave.get_peaks(), "Cannot get peaks")
        wave.fix_double_peak_by_tail()
        wave.analyse()
        r = wave.get_parameters()
        self.assertEqual(r['n_peak'], 4)
        self.assertEqual(r['max_combo_peaks'], 3)

    def test_width_distribution(self):
        df = self.data['D1']
        series = df.set_index('time')['B14']
        wave = fnc.Waveform(series)
        self.assertTrue(wave.get_peaks(), "Cannot get peaks")
        wave.analyse()
        r = wave.get_parameters()
        self.assertGreater(r['avg_shoulder_tail'], 2)

    def test_high_frequency(self):
        wave = self.data.get_waveform('high_freq')
        self.assertTrue(wave.get_peaks(), "Cannot get peaks")
        wave.analyse()
        r = wave.get_parameters()
        self.assertEqual(r['max_combo_peaks'], 1)
        self.assertEqual(r['up_length'], 0)
        self.assertFalse(r['fail_analysis'])
        self.assertTrue(r['uniform'])

    def test_one_peak(self):
        wave = self.data.get_waveform('one_peak')
        self.assertTrue(wave.get_peaks(), "Cannot get peaks")
        wave.analyse()
        with warnings.catch_warnings(record=True) as ws:
            warnings.simplefilter("always")
            r = wave.get_parameters()
            for w in ws:
                print('\n==========\n{}-{}\n{}'.format(w.filename, w.lineno, w.message))
            self.assertEqual(len(ws), 0, "Warnning has been raised {}".format(
                [x.message for x in ws]))
        self.assertEqual(r['n_peak'], 1)
        for p in r:
            self.assertFalse(np.isnan(r[p]))

    def test_uniform(self):
        wave = self.data.get_waveform('uniform')
        wave.get_peaks()
        wave.analyse()
        self.assertEqual(wave.peak_uniform_test(), 0)

    @unittest.expectedFailure
    def test_unstable_waveform(self):
        """
        Unstable means that the valleys are fluctuating during the waveform,
        so the prominences will be "inaccurate" leading to the fault double peak
        This is not easy to fix but we can try to use "QC"
        """
        wave = self.data.get_waveform('unstable')
        wave.standardise_by_filter()
        wave.get_peaks()
        wave.analyse()
        r = wave.get_parameters()
        self.assertEqual(r['max_combo_peaks'], 1)

    def test_fft_freq_ratio(self):
        wave = self.data.get_waveform('fft')
        wave.standardise_by_filter()
        wave.get_peaks()
        wave.analyse()
        r = wave.calc_fft_freq_ratio()
        self.assertAlmostEqual(r, 1.1, 1)



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="CardioWave",
    version=__import__("cdwave").__version__,
    author="Hongbin Yang",
    author_email="yanyanghong@163.com",
    description="A tool to derive parameters from waveform data for cardiotoxicity research",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/zealseeker/cardiowave",
    packages=setuptools.find_packages(),
    include_package_data=True,
    package_data={
        'cdwave': ['param_annot.json', 'parameters.csv']
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        'statsmodels',
        'tqdm',
        'joblib'
    ],
    zip_safe=False,
)
